"""
Provides functionality for interacting with ProteomeDiscoverer data.
"""

from __future__ import absolute_import, division

from collections import Counter, defaultdict
import logging
import os
import sqlite3
import xml.etree.ElementTree as ET

from . import regexes, pep_query


LOGGER = logging.getLogger("pycamv.discoverer")


def _parse_letters(letters):
    """
    Turns a string of residue letters (i.e. "STY") into a list.

    Includes special provisions for N- and C-term modifications.

    Parameters
    ----------
    letters : str

    Returns
    -------
    tuple of str
    """
    if letters in ["N-term", "C-term"]:
        return (letters,)

    return tuple(letters)


def _get_fixed_var_mods_pd21(conn):
    query = conn.cursor().execute(
        """
        SELECT
        Workflows.WorkflowXML
        FROM
        Workflows
        """
    )

    fixed_mods = []
    var_mods = []

    for xml, in query:
        root = ET.fromstring(xml)
        params = root.findall(
            "WorkflowTree/WorkflowNode/"
            "ProcessingNodeParameters/ProcessingNodeParameter"
        )
        for param in params:
            if not param.get("IsValueSet", False) or not param.text:
                continue

            if param.get("Name", "").startswith("DynMod_"):
                var_mods.append(param.text)
            elif param.get("Name", "").startswith("StaticMod_"):
                fixed_mods.append(param.text)

    return fixed_mods, var_mods


def _get_fixed_var_mods(conn):
    try:
        query = conn.cursor().execute(
            """
            SELECT
            ProcessingNodeParameters.ParameterName,
            ProcessingNodeParameters.ParameterValue
            FROM
            ProcessingNodeParameters
            """,
        )
    except sqlite3.OperationalError:
        return _get_fixed_var_mods_pd21(conn)

    fixed_mods = []
    var_mods = []

    for name, val in query:
        if name.startswith("StaticMod_"):
            fixed_mods.append(val)
        elif name.startswith("DynMod_"):
            var_mods.append(val)

    return fixed_mods, var_mods


def _find_mod(abbrev, letter, pot_mods):
    for pot_mod in pot_mods:
        if pot_mod[0] == abbrev and letter in pot_mod[1]:
            return pot_mod

    return None


def _count_mods(mod_list):
    return [
        (count, abbrev, _parse_letters(letters))
        for (abbrev, letters), count in Counter(mod_list).items()
    ]


def _get_pep_mods(conn, pep_id, pep_seq, var_mods, fixed_mods):
    aa_mods = conn.cursor().execute(
        """
        SELECT
        Peptides.SearchEngineRank,
        AminoAcidModifications.ModificationName,
        PeptidesAminoAcidModifications.Position
        FROM
        Peptides JOIN
        PeptidesAminoAcidModifications JOIN
        AminoAcidModifications
        WHERE
        Peptides.PeptideID=:pepID AND
        PeptidesAminoAcidModifications.PeptideID=:pepID AND
        PeptidesAminoAcidModifications.AminoAcidModificationID=
        AminoAcidModifications.AminoAcidModificationID
        """,
        {
            "pepID": pep_id,
        }
    )

    pep_var_mods = []
    pep_fixed_mods = []
    rank_pos = defaultdict(set)

    for rank, abbrev, pos in aa_mods:
        letter = pep_seq[pos]
        rank_pos[rank].add((pos, abbrev))

        mod = _find_mod(abbrev, letter, var_mods)

        if mod:
            pep_var_mods.append(mod)
            continue

        mod = _find_mod(abbrev, letter, fixed_mods)

        if mod:
            pep_fixed_mods.append(mod)
            continue

        # Weird ProteomeDiscoverer options, not sure what they do...
        # But we'll treat them as Ser / Asn
        LOGGER.warning(
            "Encountered X: {} - {} - {} {} ({})".format(
                pep_id, pep_seq, pos, letter, abbrev,
            )
        )

        if letter == "X" and abbrev in ["MappingS", "MappingN"]:
            continue

        raise Exception(
            "Unexpected modification: {} {}".format(letter, abbrev)
        )

    term_mods = conn.cursor().execute(
        """
        SELECT
        AminoAcidModifications.ModificationName,
        AminoAcidModifications.PositionType
        FROM
        PeptidesTerminalModifications JOIN
        AminoAcidModifications
        WHERE
        PeptidesTerminalModifications.PeptideID=:pepID AND
        PeptidesTerminalModifications.TerminalModificationID=
        AminoAcidModifications.AminoAcidModificationID
        """,
        {
            "pepID": pep_id,
        }
    )

    for abbrev, pos_type in term_mods:
        letter = "N-term" if pos_type == 1 else "C-term"
        mod = _find_mod(abbrev, letter, var_mods)

        if mod:
            pep_var_mods.append(mod)
            continue

        mod = _find_mod(abbrev, letter, fixed_mods)

        if mod:
            pep_fixed_mods.append(mod)
            continue

        raise Exception(
            "Unexpected modification: {} {}".format(letter, abbrev)
        )

    pep_var_mods = _count_mods(pep_var_mods)
    pep_fixed_mods = _count_mods(pep_fixed_mods)

    return pep_var_mods, pep_fixed_mods, rank_pos


def _get_prot_info(conn, peptide_id):
    proteins = conn.cursor().execute(
        """
        SELECT
        ProteinAnnotations.Description
        FROM
        PeptidesProteins JOIN
        ProteinAnnotations
        WHERE
        PeptidesProteins.PeptideID=:pepID AND
        ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID
        """,
        {
            "pepID": peptide_id,
        }
    )

    accessions, descriptions = [], []

    for full_prot_desc, in proteins:
        accession, prot_desc = regexes.RE_DISCOVERER_DESCRIPTION.match(
            full_prot_desc,
        ).group(1, 2)

        if not accession:
            raise Exception(
                "Unable to find accession ID for {}".format(full_prot_desc)
            )

        accessions.append(accession)
        descriptions.append(prot_desc)

    return accessions, descriptions


def _get_peptide_queries(conn, fixed_mods, var_mods):
    out = []
    # index = 0
    # scan_used = {}
    fixed_mods = [
        regexes.RE_DYN_MODS.match(i).group(3, 4)
        for i in fixed_mods
    ]
    var_mods = [
        regexes.RE_DYN_MODS.match(i).group(3, 4)
        for i in var_mods
    ]

    try:
        query = conn.cursor().execute(
            """
            SELECT
            Peptides.PeptideID,
            Peptides.Sequence,
            PeptideScores.ScoreValue,
            SpectrumHeaders.FirstScan,
            Masspeaks.Mass,
            Masspeaks.Charge,
            FileInfos.FileName
            FROM
            Peptides JOIN
            PeptideScores JOIN
            PeptidesProteins JOIN
            ProteinAnnotations JOIN
            SpectrumHeaders JOIN
            FileInfos JOIN
            Masspeaks
            WHERE
            Peptides.SearchEngineRank=1 AND
            PeptideScores.PeptideID=Peptides.PeptideID AND
            PeptidesProteins.PeptideID=Peptides.PeptideID AND
            ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID AND
            SpectrumHeaders.SpectrumID=Peptides.SpectrumID AND
            Masspeaks.MassPeakID=SpectrumHeaders.MassPeakID AND
            FileInfos.FileID=MassPeaks.FileID
            """
        )
    except sqlite3.OperationalError:
        query = conn.cursor().execute(
            """
            SELECT
            Peptides.PeptideID,
            Peptides.Sequence,
            PeptideScores.ScoreValue,
            SpectrumHeaders.FirstScan,
            Masspeaks.Mass,
            Masspeaks.Charge,
            WorkflowInputFiles.FileName
            FROM
            Peptides JOIN
            PeptideScores JOIN
            PeptidesProteins JOIN
            ProteinAnnotations JOIN
            SpectrumHeaders JOIN
            WorkflowInputFiles JOIN
            Masspeaks
            WHERE
            Peptides.SearchEngineRank=1 AND
            PeptideScores.PeptideID=Peptides.PeptideID AND
            PeptidesProteins.PeptideID=Peptides.PeptideID AND
            ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID AND
            SpectrumHeaders.SpectrumID=Peptides.SpectrumID AND
            Masspeaks.MassPeakID=SpectrumHeaders.MassPeakID AND
            WorkflowInputFiles.FileID=MassPeaks.FileID
            """
        )

    for (
        pep_id, pep_seq, pep_score,
        scan, exp_mz, exp_z, filename,
    ) in query:
        # print(pep_id, full_prot_desc, pep_seq, scan, exp_mz, exp_z, filename)
        pep_var_mods, pep_fixed_mods, rank_pos = _get_pep_mods(
            conn, pep_id, pep_seq, var_mods, fixed_mods,
        )

        accessions, descriptions = _get_prot_info(conn, pep_id)

        out.append(
            pep_query.PeptideQuery(
                accessions,
                descriptions,
                pep_id,
                filename,
                pep_score,
                exp_mz,
                exp_z,
                pep_seq,
                pep_var_mods,
                pep_fixed_mods,
                scan,
                rank_pos=rank_pos,
            )
        )

    out = sorted(out, key=lambda x: x.scan)

    return out


def read_discoverer_msf(msf_path):
    """
    Parse a ProteomeDiscoverer MSF file.

    Parameters
    ----------
    msf_path : str
        Path to MSF file.

    Returns
    -------
    fixed_mods : list of str
    var_mods : list of str
    out : list of :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    """
    LOGGER.info(
        "Loading ProteomeDiscoverer peptides from \"{}\"".format(
            os.path.basename(msf_path),
        )
    )

    with sqlite3.connect(msf_path) as conn:
        fixed_mods, var_mods = _get_fixed_var_mods(conn)
        out = _get_peptide_queries(conn, fixed_mods, var_mods)

    return fixed_mods, var_mods, out



"""
This module provides functionality for interacting with CAMV.

Currently limited to importing and outputing scan lists.
"""

# Built-ins
from __future__ import absolute_import, division

from collections import OrderedDict
from functools import partial
import logging
import multiprocessing
import os
import tempfile
import shutil

from . import compare, export, fragments, gen_sequences, scans, search
from .scan_list import load_scan_list
from .utils import LenGen


LOGGER = logging.getLogger("pycamv.validate")


def _remap_pst(pep_mods):
    return [
        (
            count,
            mod,
            letters + (
                ("Y",)
                if (mod == "Phospho" and set(letters) == set(["S", "T"])) else
                ()
            ),
        )
        for count, mod, letters in pep_mods
    ]


def _map_seq(kv, limit_comb=False):
    pep_query = kv

    gen = gen_sequences.gen_possible_seq(
        pep_query.pep_seq,
        pep_query.pep_mods,
    )

    if limit_comb:
        return (
            pep_query,
            tuple(
                seq
                for index, seq in zip(
                    range(gen_sequences.MAX_NUM_COMB),
                    gen,
                )
            )
        )
    else:
        return (
            pep_query,
            tuple(gen),
        )


def _map_frag_compare(kv):
    pep_query, scan_query, sequence, ms_two_data, ms_data = kv
    ms_two_scan = ms_two_data[pep_query.basename][pep_query.scan]
    ms_scan = ms_data[scan_query.basename][scan_query.precursor_scan]

    frag_ions = fragments.fragment_ions(
        sequence, pep_query.pep_exp_z,
        c13_num=scan_query.c13_num,
    )

    LOGGER.debug("{} - {} ions".format(pep_query.pep_seq, len(frag_ions)))

    peaks = compare.compare_spectra(
        ms_two_scan, frag_ions,
        tol=compare.COLLISION_TOLS[scan_query.collision_type],
    )

    precursor_win = scans.get_precursor_peak_window(
        scan_query, ms_scan
    )

    label_win = scans.get_label_peak_window(pep_query, ms_two_scan)

    return (pep_query, tuple(sequence), peaks, precursor_win, label_win)


def fill_map_frag_compare(
    sequence_mapping, scan_mapping,
    ms_two_data, ms_data,
    queue, cpu_count,
):
    pool = multiprocessing.Pool(
        processes=cpu_count - 1,
    )

    total_num_seq = sum(len(i) for i in sequence_mapping.values())

    try:
        peak_hits = pool.imap_unordered(
            func=_map_frag_compare,
            iterable=LenGen(
                gen=(
                    (
                        pep_query,
                        scan_mapping[pep_query],
                        sequence,
                        ms_two_data,
                        ms_data,
                    )
                    for pep_query, sequences in sequence_mapping.items()
                    for sequence in sequences
                ),
                len=total_num_seq,
            ),
        )

        for item in peak_hits:
            queue.put(item)

        pool.close()
    except:
        pool.terminate()
        raise
    finally:
        pool.join()


def validate_spectra(
    search_path,
    raw_paths=None,
    scans_path=None,
    scan_list=None,
    out_path=None,
    cpu_count=None,
    reprocess=False,
    auto_maybe=False,
):
    """
    Generate CAMV web page for validating spectra.

    Parameters
    ----------
    search_path : str
    raw_paths : list of str
    scans_path : str, optional
    scan_list : list of int, optional
    out_path : str, optional
    cpu_count : int, optional
    reprocess : bool, optional
    auto_maybe : bool, optional
    """
    if cpu_count is None:
        try:
            cpu_count = multiprocessing.cpu_count()
        except NotImplementedError:
            cpu_count = 2

    LOGGER.debug("Validating data set with {} cpus".format(cpu_count))

    if raw_paths is None:
        raw_paths = []

    if scan_list is None:
        scan_list = []

    if scans_path is not None:
        scan_list += load_scan_list(scans_path)

    if out_path is None:
        out_path = os.path.splitext(search_path)[0] + ".camv.gz"

    # Read peptide search file
    fixed_mods, var_mods, pep_queries = search.read_search_file(search_path)

    LOGGER.info("Found info for {} peptide queries".format(len(pep_queries)))

    # Optionally filter queries using a scan list
    if scan_list:
        pep_queries = [
            query
            for query in pep_queries
            if query.scan in scan_list
        ]

    # Remap pST -> pSTY
    LOGGER.info("Remapping pST -> pSTY")
    for pep_query in pep_queries:
        pep_query.pep_var_mods = _remap_pst(pep_query.pep_var_mods)
        pep_query.pep_fixed_mods = _remap_pst(pep_query.pep_fixed_mods)

    # Get scan data from RAW file
    required_raws = set(query.basename for query in pep_queries)
    base_raw_paths = [os.path.basename(path) for path in raw_paths]

    for base_raw in required_raws:
        if base_raw in base_raw_paths:
            continue

        for base_dir in [
            os.path.dirname(search_path),
            os.path.join(os.path.dirname(search_path), "..", "MS RAW")
        ]:
            local_raw_path = os.path.join(base_dir, base_raw)
            if os.path.exists(local_raw_path):
                raw_paths.append(local_raw_path)
                break
        else:
            raise Exception(
                "Unable to find {} in input RAW files: {}"
                .format(base_raw, base_raw_paths)
            )

    # Generate sequences
    LOGGER.info(
        "Generating all possible sequence-modification combinations."
    )
    pool = multiprocessing.Pool(
        processes=cpu_count,
    )
    try:
        sequence_mapping = dict(
            pool.imap_unordered(
                partial(_map_seq, limit_comb=not reprocess),
                pep_queries,
            )
        )
        pool.close()
    except:
        pool.terminate()
        raise
    finally:
        pool.join()

    total_num_seq = sum(len(i) for i in sequence_mapping.values())

    LOGGER.info(
        "Generating fragment ions for {} possible sequences."
        .format(
            total_num_seq,
        )
    )

    out_dir = tempfile.mkdtemp()

    LOGGER.info("Getting scan data.")

    scan_queries, ms_two_data, ms_data = scans.get_scan_data(
        raw_paths, pep_queries, out_dir,
    )

    LOGGER.info("Found data for {} scans".format(len(scan_queries)))

    scan_mapping = OrderedDict(
        (pep_query, scan_query)
        for pep_query, scan_query in zip(pep_queries, scan_queries)
    )

    LOGGER.info(
        (
            "Comparing predicted to actual peaks for {} spectra "
            " ({} pep-scan combinations)."
        ).format(
            len(scan_mapping),
            total_num_seq,
        )
    )

    queue = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=fill_map_frag_compare,
        args=(
            sequence_mapping,
            scan_mapping,
            ms_two_data,
            ms_data,
            queue,
            cpu_count,
        ),
    )
    process.start()

    # XXX: Determine SILAC precursor masses?

    # XXX: Remove precursor contaminated scans from validation list?

    # Check each assignment to each scan

    # Output data
    try:
        export.export_to_sql(
            os.path.splitext(out_path)[0] + ".db",
            queue, scan_mapping,
            search_path, raw_paths,
            total_num_seq=total_num_seq,
            reprocess=reprocess,
            auto_maybe=auto_maybe,
        )
    except:
        process.terminate()
        raise
    finally:
        process.join()

    LOGGER.info(
        "Exported {} total peptide-scan combinations"
        .format(total_num_seq)
    )

    LOGGER.info("Removing directory of temporary files.")

    for raw in ms_data.values():
        raw.info['fileObject'].close()
        raw.seeker.close()
    for raw in ms_two_data.values():
        raw.info['fileObject'].close()
        raw.seeker.close()

    shutil.rmtree(out_dir)
